/*
    FAMIMA CHIME
*/

#define	F_CPU	8000000UL

#include	<avr/io.h>
#include	<avr/interrupt.h>
#include	<avr/pgmspace.h>
#include	<util/delay.h>

/*****************************************************************************
    DEFINE
*****************************************************************************/
#define	SAMPLING_FREQ		20000					// [Hz]
#define SAMPLING_INTERVAL	1000000/SAMPLING_FREQ	// [us]
#define CHANNEL_MAX			2

const uint8_t	WaveForm[8] = {1,1,1,1,0,0,0,0};

typedef struct {
	uint16_t	tone_counter;
	uint16_t	tone_steptime;
	uint8_t	tone_step;
	uint8_t	envelope_volume;
	uint8_t	envelope_counter;
	uint8_t	sc_ptr;
	uint8_t	sc_wait;
} SOUND_GENERATOR;

SOUND_GENERATOR SG[CHANNEL_MAX];

/*****************************************************************************
    GPIO
*****************************************************************************/
void initGPIO()
{
	MCUCR &= ~0x10;	// Global Pull Up Enable for All Input Pins
	DDRB  = 0b00000010;	// PORT B (0:Input / 1:Output)
	PORTB = 0b11111111;	// PullUp All INput
}

/*****************************************************************************
    Timer1 for PWM
*****************************************************************************/
void initTimer1(void)
{
	PLLCSR = 0b00000010;			// PLL Enable
	_delay_us(100);
	while( (PLLCSR & 0x01) == 0 );	// Wait Phase Lock
	PLLCSR = 0b00000110;			// PCK Enable

	// TCCR1 [CTC1, PWM1A, COM1A1, COM1A0, CS13,  CS12,  CS11, CS10]
	// CTC1       = 1
	// PWM1A      = 1
	// COM1A[1:0] = 10   (Use OC1A pin, Set@00, Clear@OCR1A)
	// CS1[3:0]   = 0001 (Clock Select = PLL Clock with No PreScaling)
	TCCR1 = 0b11100001;

	GTCCR = 0b00000000;	// TSM PWM1B COM1B1 COM1B0 FOC1B FOC1A PSR1 PSR0

	OCR1A  = 0;			// PWM Pulse Width
	OCR1B  = 0;			// Not Used
	OCR1C  = 63;		// PWM Counter (from 0 to 63)
}

/*****************************************************************************
    Timer0 for Sound Generation Interrupt
*****************************************************************************/
void initTimer0(void)
{
    cli();	// Disable Interrupt

    // CTC Mode with PreScaler 1/8
    TCCR0A = 0b00000010;
    TCCR0B = 0b00000010;

    OCR0A  = F_CPU/8/SAMPLING_FREQ-1;	// Compare Match A
    OCR0B  = 0;

    TIMSK  = 0b00010000;	// Enable Timer0 Compare MatchA Interrupt

    sei();	// Enable Interrupt
}

ISR(TIMER0_COMPA_vect)
{
    static uint8_t	output = 0;
    uint8_t i;

    OCR1A = output;

    output = 0;
    for(i=0;i<CHANNEL_MAX;i++){
        SG[i].tone_counter += SAMPLING_INTERVAL*10;
        if( SG[i].tone_counter > SG[i].tone_steptime ){
            SG[i].tone_counter -= SG[i].tone_steptime;
            SG[i].tone_step = (SG[i].tone_step + 1) & 0x07;
        }
        output += ( WaveForm[SG[i].tone_step]!=0 ? SG[i].envelope_volume : 0 );
    }
}

/*****************************************************************************
    PLAY MUSIC
*****************************************************************************/
#define REST		0x7F
#define END			0xFF
#define TONE_C		0
#define TONE_CS		1
#define TONE_D		2
#define TONE_DS		3
#define TONE_E		4
#define TONE_F		5
#define TONE_FS		6
#define TONE_G		7
#define TONE_GS		8
#define TONE_A		9
#define TONE_AS		10
#define TONE_B		11

#define LENGTH_BASE	42
#define LENGTH_8	LENGTH_BASE*1-1
#define LENGTH_4	LENGTH_BASE*2-1
#define LENGTH_4P	LENGTH_BASE*3-1
#define LENGTH_2	LENGTH_BASE*4-1

const uint16_t	FreqTimeTable[24] = {
//  C,    C+,   D,    D+,   E,    F,    F+,   G,    G+,   A,    A+,   B
    4778, 4510, 4257, 4018, 3792, 3579, 3378, 3189, 3010, 2841, 2681, 2531,
    2389, 2255, 2128, 2009, 1896, 1790, 1689, 1594, 1505, 1420, 1341, 1265
};

/* !!쌠NQ̋ꂪ̂Ń`C̃fB[͕ύXĂ܂ */
/* QlɂȂǂāAeōDȂ悤ɏĂ */
const uint8_t MUSIC_DATA[2][23] PROGMEM = {
  {
    TONE_DS+12,	LENGTH_8,
    TONE_ES+12,	LENGTH_8,
    TONE_G,		LENGTH_8,
    TONE_DS+12,	LENGTH_8,
    TONE_F +12,	LENGTH_8,
    TONE_AS+12,	LENGTH_4P,
    TONE_F +12, LENGTH_8,
    TONE_G +12, LENGTH_8,
    TONE_F +12, LENGTH_8,
    TONE_AS+12,	LENGTH_8,
    TONE_DS+12,	LENGTH_2,
    END
  },
  {
    REST,		LENGTH_8,
    REST,		LENGTH_8,
    TONE_G,		LENGTH_4,
    TONE_D +12,	LENGTH_4P,
    TONE_F,		LENGTH_8,
    TONE_AS,	LENGTH_4,
    TONE_AS,	LENGTH_4,
    TONE_G,		LENGTH_2,
    END
  }
};

uint8_t playMusic(uint8_t reset)
{
    uint8_t i, rc;
    uint8_t	tone_data;
    uint8_t	wait_data;

    if(reset != 0){
        for(i=0; i<CHANNEL_MAX; i++){
            SG[i].sc_ptr  = 0;
            SG[i].sc_wait = 0;
        }
    }

    rc = 0;
    for(i=0; i<CHANNEL_MAX; i++){
        // Music Control
        if( SG[i].sc_wait > 0 ){
            SG[i].sc_wait--;
        }
        else{
            tone_data = pgm_read_byte( &MUSIC_DATA[i][SG[i].sc_ptr] );
            wait_data = pgm_read_byte( &MUSIC_DATA[i][SG[i].sc_ptr+1] );
            if( tone_data != END ){
                if( tone_data == REST ){
                    cli();
                    SG[i].tone_counter = 0;
                    SG[i].tone_steptime = FreqTimeTable[ 0 ];
                    SG[i].envelope_volume = 0;
                    SG[i].envelope_counter = 0;
                    sei();
                }
                else{
                    cli();
                    SG[i].tone_counter = 0;
                    SG[i].tone_steptime = FreqTimeTable[ tone_data ];
                    SG[i].envelope_volume = 31;
                    SG[i].envelope_counter = 8;
                    sei();
                }

                SG[i].sc_wait = wait_data;
                SG[i].sc_ptr += 2;
            }
            else{
                if( SG[i].envelope_volume ==  0){
                    rc++;
                }
            }
        }

        // Envelope Control
        if( SG[i].envelope_counter > 0 ){
            SG[i].envelope_counter--;
        }
        else{
            SG[i].envelope_counter = 8;
            if( SG[i].envelope_volume > 0 ){
                SG[i].envelope_volume--;
            }
        }
    }

    return rc;
}

void initWaveForm(void)
{
    uint8_t	i;

    for(i=0;i<CHANNEL_MAX;i++){
        SG[i].tone_counter = 0;
        SG[i].tone_steptime = 0;
        SG[i].tone_step = 0;
        SG[i].envelope_volume = 0;
        SG[i].envelope_counter = 0;
        SG[i].sc_ptr = 0;
        SG[i].sc_wait = 0;
    }
}

/*****************************************************************************
    Main
*****************************************************************************/
int main(void)
{
    uint8_t prevBtn = 0;
    uint8_t btn = 0;
    uint8_t rc;

    initGPIO();
    initWaveForm();
    initTimer1();
    initTimer0();

    while(1){
        prevBtn = btn;
        btn = ~PINB & 0x01;
        rc = playMusic( (btn^prevBtn) & btn );
        if(rc == CHANNEL_MAX){
            // Music End (enter sleep if you need)
        }
        _delay_ms(4);
    }
}
